# Copyright (c) 2012, Cloudscaling
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ast
import re

from hacking import core
import pycodestyle


"""
Guidelines for writing new hacking checks

 - Use only for Manila specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range M3xx. Find the current test with
   the highest allocated number and then pick the next value.
 - Keep the test method code in the source file ordered based
   on the M3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to manila/tests/test_hacking.py

"""

UNDERSCORE_IMPORT_FILES = []

translated_log = re.compile(
    r"(.)*LOG\."
    r"(audit|debug|error|info|warn|warning|critical|exception)"
    r"\("
    r"(_|_LE|_LI|_LW)"
    r"\(")
string_translation = re.compile(r"[^_]*_\(\s*('|\")")
underscore_import_check = re.compile(r"(.)*import _$")
underscore_import_check_multi = re.compile(r"(.)*import (.)*_, (.)*")
# We need this for cases where they have created their own _ function.
custom_underscore_check = re.compile(r"(.)*_\s*=\s*(.)*")
oslo_namespace_imports = re.compile(r"from[\s]*oslo[.](.*)")
dict_constructor_with_list_copy_re = re.compile(r".*\bdict\((\[)?(\(|\[)")
assert_no_xrange_re = re.compile(r"\s*xrange\s*\(")
assert_True = re.compile(r".*assertEqual\(True, .*\)")
no_log_warn = re.compile(r"\s*LOG.warn\(.*")
third_party_mock = re.compile(r"^import.mock")
from_third_party_mock = re.compile(r"^from.mock.import")


class BaseASTChecker(ast.NodeVisitor):
    """Provides a simple framework for writing AST-based checks.

    Subclasses should implement visit_* methods like any other AST visitor
    implementation. When they detect an error for a particular node the
    method should call ``self.add_error(offending_node)``. Details about
    where in the code the error occurred will be pulled from the node
    object.

    Subclasses should also provide a class variable named CHECK_DESC to
    be used for the human readable error message.

    """

    CHECK_DESC = 'No check message specified'

    def __init__(self, tree, filename):
        """This object is created automatically by pep8.

        :param tree: an AST tree
        :param filename: name of the file being analyzed
                         (ignored by our checks)
        """
        self._tree = tree
        self._errors = []

    def run(self):
        """Called automatically by pep8."""
        self.visit(self._tree)
        return self._errors

    def add_error(self, node, message=None):
        """Add an error caused by a node to the list of errors for pep8."""
        message = message or self.CHECK_DESC
        error = (node.lineno, node.col_offset, message, self.__class__)
        self._errors.append(error)

    def _check_call_names(self, call_node, names):
        if isinstance(call_node, ast.Call):
            if isinstance(call_node.func, ast.Name):
                if call_node.func.id in names:
                    return True
        return False


@core.flake8ext
def no_translate_logs(logical_line):
    if translated_log.match(logical_line):
        yield(0, "M359 Don't translate log messages!")


class CheckLoggingFormatArgs(BaseASTChecker):
    """Check for improper use of logging format arguments.

    LOG.debug("Volume %s caught fire and is at %d degrees C and climbing.",
              ('volume1', 500))

    The format arguments should not be a tuple as it is easy to miss.

    """

    name = "check_logging_format_args"
    version = "1.0"
    CHECK_DESC = 'M310 Log method arguments should not be a tuple.'
    LOG_METHODS = [
        'debug', 'info',
        'warn', 'warning',
        'error', 'exception',
        'critical', 'fatal',
        'trace', 'log'
    ]

    def _find_name(self, node):
        """Return the fully qualified name or a Name or Attribute."""
        if isinstance(node, ast.Name):
            return node.id
        elif (isinstance(node, ast.Attribute)
                and isinstance(node.value, (ast.Name, ast.Attribute))):
            method_name = node.attr
            obj_name = self._find_name(node.value)
            if obj_name is None:
                return None
            return obj_name + '.' + method_name
        elif isinstance(node, str):
            return node
        else:  # could be Subscript, Call or many more
            return None

    def visit_Call(self, node):
        """Look for the 'LOG.*' calls."""
        # extract the obj_name and method_name
        if isinstance(node.func, ast.Attribute):
            obj_name = self._find_name(node.func.value)
            if isinstance(node.func.value, ast.Name):
                method_name = node.func.attr
            elif isinstance(node.func.value, ast.Attribute):
                obj_name = self._find_name(node.func.value)
                method_name = node.func.attr
            else:  # could be Subscript, Call or many more
                return super(CheckLoggingFormatArgs, self).generic_visit(node)

            # obj must be a logger instance and method must be a log helper
            if (obj_name != 'LOG'
                    or method_name not in self.LOG_METHODS):
                return super(CheckLoggingFormatArgs, self).generic_visit(node)

            # the call must have arguments
            if not len(node.args):
                return super(CheckLoggingFormatArgs, self).generic_visit(node)

            # any argument should not be a tuple
            for arg in node.args:
                if isinstance(arg, ast.Tuple):
                    self.add_error(arg)

        return super(CheckLoggingFormatArgs, self).generic_visit(node)


@core.flake8ext
def check_explicit_underscore_import(logical_line, filename):
    """Check for explicit import of the _ function

    We need to ensure that any files that are using the _() function
    to translate logs are explicitly importing the _ function.  We
    can't trust unit test to catch whether the import has been
    added so we need to check for it here.
    """

    # Build a list of the files that have _ imported.  No further
    # checking needed once it is found.
    if filename in UNDERSCORE_IMPORT_FILES:
        pass
    elif (underscore_import_check.match(logical_line) or
          underscore_import_check_multi.match(logical_line) or
          custom_underscore_check.match(logical_line)):
        UNDERSCORE_IMPORT_FILES.append(filename)
    elif string_translation.match(logical_line):
        yield(0, "M323: Found use of _() without explicit import of _ !")


class CheckForTransAdd(BaseASTChecker):
    """Checks for the use of concatenation on a translated string.

    Translations should not be concatenated with other strings, but
    should instead include the string being added to the translated
    string to give the translators the most information.
    """

    name = "check_for_trans_add"
    version = "1.0"
    CHECK_DESC = ('M326 Translated messages cannot be concatenated.  '
                  'String should be included in translated message.')

    TRANS_FUNC = ['_', '_LI', '_LW', '_LE', '_LC']

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Add):
            if self._check_call_names(node.left, self.TRANS_FUNC):
                self.add_error(node.left)
            elif self._check_call_names(node.right, self.TRANS_FUNC):
                self.add_error(node.right)
        super(CheckForTransAdd, self).generic_visit(node)


@core.flake8ext
def check_oslo_namespace_imports(physical_line, logical_line, filename):
    if pycodestyle.noqa(physical_line):
        return
    if re.match(oslo_namespace_imports, logical_line):
        msg = ("M333: '%s' must be used instead of '%s'.") % (
            logical_line.replace('oslo.', 'oslo_'),
            logical_line)
        yield(0, msg)


@core.flake8ext
def dict_constructor_with_list_copy(logical_line):
    msg = ("M336: Must use a dict comprehension instead of a dict constructor"
           " with a sequence of key-value pairs."
           )
    if dict_constructor_with_list_copy_re.match(logical_line):
        yield (0, msg)


@core.flake8ext
def no_xrange(logical_line):
    if assert_no_xrange_re.match(logical_line):
        yield(0, "M337: Do not use xrange().")


@core.flake8ext
def validate_assertTrue(logical_line):
    if re.match(assert_True, logical_line):
        msg = ("M313: Unit tests should use assertTrue(value) instead"
               " of using assertEqual(True, value).")
        yield(0, msg)


@core.flake8ext
def check_uuid4(logical_line):
    """Generating UUID

    Use oslo_utils.uuidutils to generate UUID instead of uuid4().

    M354
    """

    msg = ("M354: Use oslo_utils.uuidutils to generate UUID instead "
           "of uuid4().")

    if "uuid4()." in logical_line:
        return

    if "uuid4()" in logical_line:
        yield (0, msg)


@core.flake8ext
def no_log_warn_check(logical_line):
    """Disallow 'LOG.warn'

    Deprecated LOG.warn(), instead use LOG.warning
    ://bugs.launchpad.net/manila/+bug/1508442

    M338
    """
    msg = ("M338: LOG.warn is deprecated, use LOG.warning.")
    if re.match(no_log_warn, logical_line):
        yield(0, msg)


@core.flake8ext
def no_third_party_mock(logical_line):
    # We should only use unittest.mock, not the third party mock library that
    # was needed for py2 support.
    if (re.match(third_party_mock, logical_line) or
            re.match(from_third_party_mock, logical_line)):
        msg = ('M339: Unit tests should use the standard library "mock" '
               'module, not the third party mock library.')
        yield(0, msg)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Quotas for shares."""

import datetime

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils
from oslo_utils import timeutils
import six

from manila import db
from manila import exception

LOG = log.getLogger(__name__)

QUOTA_GROUP = 'quota'

quota_opts = [
    cfg.IntOpt('shares',
               default=50,
               help='Number of shares allowed per project.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_shares'),
    cfg.IntOpt('snapshots',
               default=50,
               help='Number of share snapshots allowed per project.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_snapshots'),
    cfg.IntOpt('gigabytes',
               default=1000,
               help='Number of share gigabytes allowed per project.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_gigabytes'),
    cfg.IntOpt('per_share_gigabytes',
               default=-1,
               help='Max size allowed per share, in gigabytes.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_per_share_gigabytes'),
    cfg.IntOpt('snapshot_gigabytes',
               default=1000,
               help='Number of snapshot gigabytes allowed per project.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_snapshot_gigabytes'),
    cfg.IntOpt('share_networks',
               default=10,
               help='Number of share-networks allowed per project.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_share_networks'),
    cfg.IntOpt('share_replicas',
               default=100,
               help='Number of share-replicas allowed per project.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_share_replicas'),
    cfg.IntOpt('replica_gigabytes',
               default=1000,
               help='Number of replica gigabytes allowed per project.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_replica_gigabytes'),
    cfg.IntOpt('share_groups',
               default=50,
               help='Number of share groups allowed.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_share_groups'),
    cfg.IntOpt('share_group_snapshots',
               default=50,
               help='Number of share group snapshots allowed.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_share_group_snapshots'),
    cfg.IntOpt('reservation_expire',
               default=86400,
               help='Number of seconds until a reservation expires.',
               deprecated_group='DEFAULT'),
    cfg.IntOpt('until_refresh',
               default=0,
               help='Count of reservations until usage is refreshed.',
               deprecated_group='DEFAULT'),
    cfg.IntOpt('max_age',
               default=0,
               help='Number of seconds between subsequent usage refreshes.',
               deprecated_group='DEFAULT'),
    cfg.StrOpt('driver',
               default='manila.quota.DbQuotaDriver',
               help='Default driver to use for quota checks.',
               deprecated_group='DEFAULT',
               deprecated_name='quota_driver'), ]

CONF = cfg.CONF
CONF.register_opts(quota_opts, QUOTA_GROUP)


class DbQuotaDriver(object):
    """Database Quota driver.

    Driver to perform necessary checks to enforce quotas and obtain
    quota information.  The default driver utilizes the local
    database.
    """

    def get_by_class(self, context, quota_class, resource):
        """Get a specific quota by quota class."""

        return db.quota_class_get(context, quota_class, resource)

    def get_defaults(self, context, resources):
        """Given a list of resources, retrieve the default quotas.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        """

        quotas = {}
        default_quotas = db.quota_class_get_default(context)
        for resource in resources.values():
            quotas[resource.name] = default_quotas.get(resource.name,
                                                       resource.default)
        return quotas

    def get_class_quotas(self, context, resources, quota_class,
                         defaults=True):
        """Retrieve quotas for a quota class.

        Given a list of resources, retrieve the quotas for the given
        quota class.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param quota_class: The name of the quota class to return
                            quotas for.
        :param defaults: If True, the default value will be reported
                         if there is no specific value for the
                         resource.
        """

        quotas = {}
        class_quotas = db.quota_class_get_all_by_name(context, quota_class)
        for resource in resources.values():
            if defaults or resource.name in class_quotas:
                quotas[resource.name] = class_quotas.get(resource.name,
                                                         resource.default)

        return quotas

    def _process_quotas(self, context, resources, project_id, quotas,
                        quota_class=None, defaults=True, usages=None,
                        remains=False):
        modified_quotas = {}
        # Get the quotas for the appropriate class.  If the project ID
        # matches the one in the context, we use the quota_class from
        # the context, otherwise, we use the provided quota_class (if
        # any)
        if project_id == context.project_id:
            quota_class = context.quota_class
        if quota_class:
            class_quotas = db.quota_class_get_all_by_name(context, quota_class)
        else:
            class_quotas = {}

        default_quotas = self.get_defaults(context, resources)

        for resource in resources.values():
            # Omit default/quota class values
            if not defaults and resource.name not in quotas:
                continue

            limit = quotas.get(
                resource.name,
                class_quotas.get(resource.name, default_quotas[resource.name]))
            modified_quotas[resource.name] = dict(limit=limit)

            # Include usages if desired.  This is optional because one
            # internal consumer of this interface wants to access the
            # usages directly from inside a transaction.
            if usages:
                usage = usages.get(resource.name, {})
                modified_quotas[resource.name].update(
                    in_use=usage.get('in_use', 0),
                    reserved=usage.get('reserved', 0),
                )
            # Initialize remains quotas.
            if remains:
                modified_quotas[resource.name].update(remains=limit)

        if remains:
            all_quotas = db.quota_get_all(context, project_id)
            for quota in all_quotas:
                if quota.resource in modified_quotas:
                    modified_quotas[quota.resource]['remains'] -= (
                        quota.hard_limit)

        return modified_quotas

    def get_project_quotas(self, context, resources, project_id,
                           quota_class=None, defaults=True,
                           usages=True, remains=False):
        """Retrieve quotas for project.

        Given a list of resources, retrieve the quotas for the given
        project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.  It
                            will be ignored if project_id ==
                            context.project_id.
        :param defaults: If True, the quota class value (or the
                         default value, if there is no value from the
                         quota class) will be reported if there is no
                         specific value for the resource.
        :param usages: If True, the current in_use and reserved counts
                       will also be returned.
        :param remains: If True, the current remains of the project will
                        will be returned.
        """
        project_quotas = db.quota_get_all_by_project(context, project_id)
        project_usages = None
        if usages:
            project_usages = db.quota_usage_get_all_by_project(context,
                                                               project_id)
        return self._process_quotas(context, resources, project_id,
                                    project_quotas, quota_class,
                                    defaults=defaults, usages=project_usages,
                                    remains=remains)

    def get_user_quotas(self, context, resources, project_id, user_id,
                        quota_class=None, defaults=True, usages=True):
        """Retrieve quotas for user and project.

        Given a list of resources, retrieve the quotas for the given
        user and project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.  It
                            will be ignored if project_id ==
                            context.project_id.
        :param defaults: If True, the quota class value (or the
                         default value, if there is no value from the
                         quota class) will be reported if there is no
                         specific value for the resource.
        :param usages: If True, the current in_use and reserved counts
                       will also be returned.
        """
        user_quotas = db.quota_get_all_by_project_and_user(
            context, project_id, user_id)
        # Use the project quota for default user quota.
        proj_quotas = db.quota_get_all_by_project(context, project_id)
        for key, value in proj_quotas.items():
            if key not in user_quotas.keys():
                user_quotas[key] = value
        user_usages = None
        if usages:
            user_usages = db.quota_usage_get_all_by_project_and_user(
                context, project_id, user_id)
        return self._process_quotas(context, resources, project_id,
                                    user_quotas, quota_class,
                                    defaults=defaults, usages=user_usages)

    def get_share_type_quotas(self, context, resources, project_id,
                              share_type_id, quota_class=None, defaults=True,
                              usages=True):
        """Retrieve quotas for share_type and project.

        Given a list of resources, retrieve the quotas for the given
        share_type and project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The UUID of the project to return quotas for.
        :param share_type: UUID/name of a share type to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.  It
                            will be ignored if project_id ==
                            context.project_id.
        :param defaults: If True, the quota class value (or the
                         default value, if there is no value from the
                         quota class) will be reported if there is no
                         specific value for the resource.
        :param usages: If True, the current in_use and reserved counts
                       will also be returned.
        """
        st_quotas = db.quota_get_all_by_project_and_share_type(
            context, project_id, share_type_id)
        # Use the project quota for default share_type quota.
        project_quotas = db.quota_get_all_by_project(context, project_id)
        for key, value in project_quotas.items():
            if key not in st_quotas.keys():
                st_quotas[key] = value
        st_usages = None
        if usages:
            st_usages = db.quota_usage_get_all_by_project_and_share_type(
                context, project_id, share_type_id)
        return self._process_quotas(
            context, resources, project_id, st_quotas, quota_class,
            defaults=defaults, usages=st_usages)

    def get_settable_quotas(self, context, resources, project_id,
                            user_id=None, share_type_id=None):
        """Retrieve range of settable quotas.

        Given a list of resources, retrieve the range of settable quotas for
        the given user or project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        :param share_type_id: The UUID of the share_type to return quotas for.
        """
        settable_quotas = {}
        project_quotas = self.get_project_quotas(
            context, resources, project_id, remains=True)
        if user_id or share_type_id:
            if user_id:
                subquotas = self.get_user_quotas(
                    context, resources, project_id, user_id)
            else:
                subquotas = self.get_share_type_quotas(
                    context, resources, project_id, share_type_id)
            for key, value in subquotas.items():
                settable_quotas[key] = {
                    "minimum": value['in_use'] + value['reserved'],
                    "maximum": project_quotas[key]["limit"],
                }
        else:
            for key, value in project_quotas.items():
                minimum = max(
                    int(value['limit'] - value['remains']),
                    int(value['in_use'] + value['reserved'])
                )
                settable_quotas[key] = {"minimum": minimum, "maximum": -1}
        return settable_quotas

    def _get_quotas(self, context, resources, keys, has_sync, project_id=None,
                    user_id=None, share_type_id=None):
        """Retrieve quotas for a resource.

        A helper method which retrieves the quotas for the specific
        resources identified by keys, and which apply to the current
        context.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param keys: A list of the desired quotas to retrieve.
        :param has_sync: If True, indicates that the resource must
                         have a sync attribute; if False, indicates
                         that the resource must NOT have a sync
                         attribute.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        :param user_id: Specify the user_id if current context
                        is admin and admin wants to impact on
                        common user. (Special case: user operates on
                        resource, owned/created by different user)
        """

        # Filter resources
        if has_sync:
            sync_filt = lambda x: hasattr(x, 'sync')  # noqa: E731
        else:
            sync_filt = lambda x: not hasattr(x, 'sync')  # noqa: E731
        desired = set(keys)
        sub_resources = {k: v for k, v in resources.items()
                         if k in desired and sync_filt(v)}

        # Make sure we accounted for all of them...
        if len(keys) != len(sub_resources):
            unknown = desired - set(sub_resources.keys())
            raise exception.QuotaResourceUnknown(unknown=sorted(unknown))

        if user_id:
            # Grab and return the quotas (without usages)
            quotas = self.get_user_quotas(context, sub_resources,
                                          project_id, user_id,
                                          context.quota_class, usages=False)
        elif share_type_id:
            # Grab and return the quotas (without usages)
            quotas = self.get_share_type_quotas(
                context, sub_resources, project_id, share_type_id,
                context.quota_class, usages=False)
        else:
            # Grab and return the quotas (without usages)
            quotas = self.get_project_quotas(context, sub_resources,
                                             project_id,
                                             context.quota_class,
                                             usages=False)

        return {k: v['limit'] for k, v in quotas.items()}

    def limit_check(self, context, resources, values, project_id=None):
        """Check simple quota limits.

        For limits--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit
        resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param values: A dictionary of the values to check against the
                       quota.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        """

        # Ensure no value is less than zero
        unders = [key for key, val in values.items() if val < 0]
        if unders:
            raise exception.InvalidQuotaValue(unders=sorted(unders))

        # If project_id is None, then we use the project_id in context
        if project_id is None:
            project_id = context.project_id

        quotas = self._get_quotas(context, resources, values.keys(),
                                  has_sync=False, project_id=project_id)
        # Check the quotas and construct a list of the resources that
        # would be put over limit by the desired values
        overs = [key for key, val in values.items()
                 if quotas[key] >= 0 and quotas[key] < val]
        if overs:
            raise exception.OverQuota(overs=sorted(overs), quotas=quotas,
                                      usages={})

    def reserve(self, context, resources, deltas, expire=None,
                project_id=None, user_id=None, share_type_id=None,
                overquota_allowed=False):
        """Check quotas and reserve resources.

        For counting quotas--those quotas for which there is a usage
        synchronization function--this method checks quotas against
        current usage and the desired deltas.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it does not have a usage
        synchronization function.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns a
        list of reservation UUIDs which were created.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param deltas: A dictionary of the proposed delta changes.
        :param expire: An optional parameter specifying an expiration
                       time for the reservations.  If it is a simple
                       number, it is interpreted as a number of
                       seconds and added to the current time; if it is
                       a datetime.timedelta object, it will also be
                       added to the current time.  A datetime.datetime
                       object will be interpreted as the absolute
                       expiration time.  If None is specified, the
                       default expiration time set by
                       --default-reservation-expire will be used (this
                       value will be treated as a number of seconds).
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        :param user_id: Specify the user_id if current context
                        is admin and admin wants to impact on
                        common user. (Special case: user operates on
                        resource, owned/created by different user)
        """

        # Set up the reservation expiration
        if expire is None:
            expire = CONF.quota.reservation_expire
        if isinstance(expire, six.integer_types):
            expire = datetime.timedelta(seconds=expire)
        if isinstance(expire, datetime.timedelta):
            expire = timeutils.utcnow() + expire
        if not isinstance(expire, datetime.datetime):
            raise exception.InvalidReservationExpiration(expire=expire)

        # If project_id is None, then we use the project_id in context
        if project_id is None:
            project_id = context.project_id
        # If user_id is None, then we use the user_id in context
        if user_id is None:
            user_id = context.user_id

        # Get the applicable quotas.
        # NOTE(Vek): We're not worried about races at this point.
        #            Yes, the admin may be in the process of reducing
        #            quotas, but that's a pretty rare thing.
        quotas = self._get_quotas(
            context, resources, deltas, has_sync=True, project_id=project_id)
        user_quotas = self._get_quotas(
            context, resources, deltas, has_sync=True,
            project_id=project_id, user_id=user_id)
        if share_type_id:
            share_type_quotas = self._get_quotas(
                context, resources, deltas, has_sync=True,
                project_id=project_id, share_type_id=share_type_id)
        else:
            share_type_quotas = {}

        # NOTE(Vek): Most of the work here has to be done in the DB
        #            API, because we have to do it in a transaction,
        #            which means access to the session.  Since the
        #            session isn't available outside the DBAPI, we
        #            have to do the work there.
        return db.quota_reserve(
            context, resources, quotas, user_quotas, share_type_quotas,
            deltas, expire, CONF.quota.until_refresh, CONF.quota.max_age,
            project_id=project_id, user_id=user_id,
            share_type_id=share_type_id, overquota_allowed=overquota_allowed)

    def commit(self, context, reservations, project_id=None, user_id=None,
               share_type_id=None):
        """Commit reservations.

        :param context: The request context, for access checks.
        :param reservations: A list of the reservation UUIDs, as
                             returned by the reserve() method.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        :param user_id: Specify the user_id if current context
                        is admin and admin wants to impact on
                        common user. (Special case: user operates on
                        resource, owned/created by different user)
        """
        # If project_id is None, then we use the project_id in context
        if project_id is None:
            project_id = context.project_id
        # If user_id is None, then we use the user_id in context
        if user_id is None:
            user_id = context.user_id

        db.reservation_commit(
            context, reservations, project_id=project_id, user_id=user_id,
            share_type_id=share_type_id)

    def rollback(self, context, reservations, project_id=None, user_id=None,
                 share_type_id=None):
        """Roll back reservations.

        :param context: The request context, for access checks.
        :param reservations: A list of the reservation UUIDs, as
                             returned by the reserve() method.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        :param user_id: Specify the user_id if current context
                        is admin and admin wants to impact on
                        common user. (Special case: user operates on
                        resource, owned/created by different user)
        """
        # If project_id is None, then we use the project_id in context
        if project_id is None:
            project_id = context.project_id
        # If user_id is None, then we use the user_id in context
        if user_id is None:
            user_id = context.user_id

        db.reservation_rollback(
            context, reservations, project_id=project_id, user_id=user_id,
            share_type_id=share_type_id)

    def usage_reset(self, context, resources):
        """Reset usage records.

        Reset the usage records for a particular user on a list of
        resources.  This will force that user's usage records to be
        refreshed the next time a reservation is made.

        Note: this does not affect the currently outstanding
        reservations the user has; those reservations must be
        committed or rolled back (or expired).

        :param context: The request context, for access checks.
        :param resources: A list of the resource names for which the
                          usage must be reset.
        """

        # We need an elevated context for the calls to
        # quota_usage_update()
        elevated = context.elevated()

        for resource in resources:
            try:
                # Reset the usage to -1, which will force it to be
                # refreshed
                db.quota_usage_update(elevated, context.project_id,
                                      context.user_id,
                                      resource, in_use=-1)
            except exception.QuotaUsageNotFound:
                # That means it'll be refreshed anyway
                pass

    def destroy_all_by_project(self, context, project_id):
        """Destroy metadata associated with a project.

        Destroy all quotas, usages, and reservations associated with a
        project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project being deleted.
        """

        db.quota_destroy_all_by_project(context, project_id)

    def destroy_all_by_project_and_user(self, context, project_id, user_id):
        """Destroy metadata associated with a project and user.

        Destroy all quotas, usages, and reservations associated with a
        project and user.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project being deleted.
        :param user_id: The ID of the user being deleted.
        """

        db.quota_destroy_all_by_project_and_user(context, project_id, user_id)

    def destroy_all_by_project_and_share_type(self, context, project_id,
                                              share_type_id):
        """Destroy metadata associated with a project and share_type.

        Destroy all quotas, usages, and reservations associated with a
        project and share_type.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project.
        :param share_type_id: The UUID of the share type.
        """

        db.quota_destroy_all_by_share_type(context, share_type_id,
                                           project_id=project_id)

    def expire(self, context):
        """Expire reservations.

        Explores all currently existing reservations and rolls back
        any that have expired.

        :param context: The request context, for access checks.
        """

        db.reservation_expire(context)


class BaseResource(object):
    """Describe a single resource for quota checking."""

    def __init__(self, name, flag=None):
        """Initializes a Resource.

        :param name: The name of the resource, i.e., "shares".
        :param flag: The name of the flag or configuration option
                     which specifies the default value of the quota
                     for this resource.
        """

        self.name = name
        self.flag = flag

    @property
    def default(self):
        """Return the default value of the quota."""

        return CONF.quota[self.flag] if self.flag else -1


class ReservableResource(BaseResource):
    """Describe a reservable resource."""

    def __init__(self, name, sync, flag=None):
        """Initializes a ReservableResource.

        Reservable resources are those resources which directly
        correspond to objects in the database, i.e., shares, gigabytes,
        etc.  A ReservableResource must be constructed with a usage
        synchronization function, which will be called to determine the
        current counts of one or more resources.

        The usage synchronization function will be passed three
        arguments: an admin context, the project ID, and an opaque
        session object, which should in turn be passed to the
        underlying database function.  Synchronization functions
        should return a dictionary mapping resource names to the
        current in_use count for those resources; more than one
        resource and resource count may be returned.  Note that
        synchronization functions may be associated with more than one
        ReservableResource.

        :param name: The name of the resource, i.e., "shares".
        :param sync: A callable which returns a dictionary to
                     resynchronize the in_use count for one or more
                     resources, as described above.
        :param flag: The name of the flag or configuration option
                     which specifies the default value of the quota
                     for this resource.
        """

        super(ReservableResource, self).__init__(name, flag=flag)
        if sync:
            self.sync = sync


class AbsoluteResource(BaseResource):
    """Describe a non-reservable resource."""

    pass


class CountableResource(AbsoluteResource):
    """Describe a countable resource.

    Describe a resource where the counts aren't based solely on the
    project ID.
    """

    def __init__(self, name, count, flag=None):
        """Initializes a CountableResource.

        Countable resources are those resources which directly
        correspond to objects in the database, i.e., shares, gigabytes,
        etc., but for which a count by project ID is inappropriate.  A
        CountableResource must be constructed with a counting
        function, which will be called to determine the current counts
        of the resource.

        The counting function will be passed the context, along with
        the extra positional and keyword arguments that are passed to
        Quota.count().  It should return an integer specifying the
        count.

        Note that this counting is not performed in a transaction-safe
        manner.  This resource class is a temporary measure to provide
        required functionality, until a better approach to solving
        this problem can be evolved.

        :param name: The name of the resource, i.e., "shares".
        :param count: A callable which returns the count of the
                      resource.  The arguments passed are as described
                      above.
        :param flag: The name of the flag or configuration option
                     which specifies the default value of the quota
                     for this resource.
        """

        super(CountableResource, self).__init__(name, flag=flag)
        self.count = count


class QuotaEngine(object):
    """Represent the set of recognized quotas."""

    def __init__(self, quota_driver_class=None):
        """Initialize a Quota object."""
        self._resources = {}
        self._driver_cls = quota_driver_class
        self.__driver = None

    @property
    def _driver(self):
        if self.__driver:
            return self.__driver
        if not self._driver_cls:
            self._driver_cls = CONF.quota.driver
        if isinstance(self._driver_cls, six.string_types):
            self._driver_cls = importutils.import_object(self._driver_cls)
        self.__driver = self._driver_cls
        return self.__driver

    def __contains__(self, resource):
        return resource in self._resources

    def register_resource(self, resource):
        """Register a resource."""

        self._resources[resource.name] = resource

    def register_resources(self, resources):
        """Register a list of resources."""

        for resource in resources:
            self.register_resource(resource)

    def get_by_class(self, context, quota_class, resource):
        """Get a specific quota by quota class."""

        return self._driver.get_by_class(context, quota_class, resource)

    def get_defaults(self, context):
        """Retrieve the default quotas.

        :param context: The request context, for access checks.
        """

        return self._driver.get_defaults(context, self._resources)

    def get_class_quotas(self, context, quota_class, defaults=True):
        """Retrieve the quotas for the given quota class.

        :param context: The request context, for access checks.
        :param quota_class: The name of the quota class to return
                            quotas for.
        :param defaults: If True, the default value will be reported
                         if there is no specific value for the
                         resource.
        """

        return self._driver.get_class_quotas(context, self._resources,
                                             quota_class, defaults=defaults)

    def get_user_quotas(self, context, project_id, user_id, quota_class=None,
                        defaults=True, usages=True):
        """Retrieve the quotas for the given user and project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.
        :param defaults: If True, the quota class value (or the
                         default value, if there is no value from the
                         quota class) will be reported if there is no
                         specific value for the resource.
        :param usages: If True, the current in_use and reserved counts
                       will also be returned.
        """

        return self._driver.get_user_quotas(context, self._resources,
                                            project_id, user_id,
                                            quota_class=quota_class,
                                            defaults=defaults,
                                            usages=usages)

    def get_share_type_quotas(self, context, project_id, share_type_id,
                              quota_class=None, defaults=True, usages=True):
        """Retrieve the quotas for the given user and project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to return quotas for.
        :param share_type_id: The UUID of the user to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.
        :param defaults: If True, the quota class value (or the
                         default value, if there is no value from the
                         quota class) will be reported if there is no
                         specific value for the resource.
        :param usages: If True, the current in_use and reserved counts
                       will also be returned.
        """

        return self._driver.get_share_type_quotas(
            context, self._resources, project_id, share_type_id,
            quota_class=quota_class, defaults=defaults, usages=usages)

    def get_project_quotas(self, context, project_id, quota_class=None,
                           defaults=True, usages=True, remains=False):
        """Retrieve the quotas for the given project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to return quotas for.
        :param quota_class: If project_id != context.project_id, the
                            quota class cannot be determined.  This
                            parameter allows it to be specified.
        :param defaults: If True, the quota class value (or the
                         default value, if there is no value from the
                         quota class) will be reported if there is no
                         specific value for the resource.
        :param usages: If True, the current in_use and reserved counts
                       will also be returned.
        :param remains: If True, the current remains of the project will
                        will be returned.
        """

        return self._driver.get_project_quotas(context, self._resources,
                                               project_id,
                                               quota_class=quota_class,
                                               defaults=defaults,
                                               usages=usages,
                                               remains=remains)

    def get_settable_quotas(self, context, project_id, user_id=None,
                            share_type_id=None):
        """Get settable quotas.

        Given a list of resources, retrieve the range of settable quotas for
        the given user or project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resources.
        :param project_id: The ID of the project to return quotas for.
        :param user_id: The ID of the user to return quotas for.
        :param share_type_id: The UUID of the share_type to return quotas for.
        """

        return self._driver.get_settable_quotas(
            context, self._resources, project_id, user_id=user_id,
            share_type_id=share_type_id)

    def count(self, context, resource, *args, **kwargs):
        """Count a resource.

        For countable resources, invokes the count() function and
        returns its result.  Arguments following the context and
        resource are passed directly to the count function declared by
        the resource.

        :param context: The request context, for access checks.
        :param resource: The name of the resource, as a string.
        """

        # Get the resource
        res = self._resources.get(resource)
        if not res or not hasattr(res, 'count'):
            raise exception.QuotaResourceUnknown(unknown=[resource])

        return res.count(context, *args, **kwargs)

    def limit_check(self, context, project_id=None, **values):
        """Check simple quota limits.

        For limits--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction.  The
        values to check are given as keyword arguments, where the key
        identifies the specific quota limit to check, and the value is
        the proposed value.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it is not a simple limit
        resource.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        """

        return self._driver.limit_check(context, self._resources, values,
                                        project_id=project_id)

    def reserve(self, context, expire=None, project_id=None, user_id=None,
                share_type_id=None, overquota_allowed=False, **deltas):
        """Check quotas and reserve resources.

        For counting quotas--those quotas for which there is a usage
        synchronization function--this method checks quotas against
        current usage and the desired deltas.  The deltas are given as
        keyword arguments, and current usage and other reservations
        are factored into the quota check.

        This method will raise a QuotaResourceUnknown exception if a
        given resource is unknown or if it does not have a usage
        synchronization function.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns a
        list of reservation UUIDs which were created.

        :param context: The request context, for access checks.
        :param expire: An optional parameter specifying an expiration
                       time for the reservations.  If it is a simple
                       number, it is interpreted as a number of
                       seconds and added to the current time; if it is
                       a datetime.timedelta object, it will also be
                       added to the current time.  A datetime.datetime
                       object will be interpreted as the absolute
                       expiration time.  If None is specified, the
                       default expiration time set by
                       --default-reservation-expire will be used (this
                       value will be treated as a number of seconds).
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        """

        reservations = self._driver.reserve(
            context, self._resources, deltas,
            expire=expire,
            project_id=project_id,
            user_id=user_id,
            share_type_id=share_type_id,
            overquota_allowed=overquota_allowed
        )

        LOG.debug("Created reservations %s", reservations)

        return reservations

    def commit(self, context, reservations, project_id=None, user_id=None,
               share_type_id=None):
        """Commit reservations.

        :param context: The request context, for access checks.
        :param reservations: A list of the reservation UUIDs, as
                             returned by the reserve() method.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        """

        try:
            self._driver.commit(
                context, reservations, project_id=project_id,
                user_id=user_id, share_type_id=share_type_id)
        except Exception:
            # NOTE(Vek): Ignoring exceptions here is safe, because the
            # usage resynchronization and the reservation expiration
            # mechanisms will resolve the issue.  The exception is
            # logged, however, because this is less than optimal.
            LOG.exception("Failed to commit reservations %s",
                          reservations)
            return
        LOG.debug("Committed reservations %s", reservations)

    def rollback(self, context, reservations, project_id=None, user_id=None,
                 share_type_id=None):
        """Roll back reservations.

        :param context: The request context, for access checks.
        :param reservations: A list of the reservation UUIDs, as
                             returned by the reserve() method.
        :param project_id: Specify the project_id if current context
                           is admin and admin wants to impact on
                           common user's tenant.
        """

        try:
            self._driver.rollback(
                context, reservations, project_id=project_id,
                user_id=user_id, share_type_id=share_type_id)
        except Exception:
            # NOTE(Vek): Ignoring exceptions here is safe, because the
            # usage resynchronization and the reservation expiration
            # mechanisms will resolve the issue.  The exception is
            # logged, however, because this is less than optimal.
            LOG.exception("Failed to roll back reservations %s",
                          reservations)
            return
        LOG.debug("Rolled back reservations %s", reservations)

    def usage_reset(self, context, resources):
        """Reset usage records.

        Reset the usage records for a particular user on a list of
        resources.  This will force that user's usage records to be
        refreshed the next time a reservation is made.

        Note: this does not affect the currently outstanding
        reservations the user has; those reservations must be
        committed or rolled back (or expired).

        :param context: The request context, for access checks.
        :param resources: A list of the resource names for which the
                          usage must be reset.
        """

        self._driver.usage_reset(context, resources)

    def destroy_all_by_project_and_user(self, context, project_id, user_id):
        """Destroy metadata associated with a project and user.

        Destroy all quotas, usages, and reservations associated with a
        project and user.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project being deleted.
        :param user_id: The ID of the user being deleted.
        """

        self._driver.destroy_all_by_project_and_user(context,
                                                     project_id, user_id)

    def destroy_all_by_project_and_share_type(self, context, project_id,
                                              share_type_id):
        """Destroy metadata associated with a project and share_type.

        Destroy all quotas, usages, and reservations associated with a
        project and share_type.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project.
        :param share_type_id: The UUID of the share_type.
        """

        self._driver.destroy_all_by_project_and_share_type(
            context, project_id, share_type_id)

    def destroy_all_by_project(self, context, project_id):
        """Destroy metadata associated with a project.

        Destroy all quotas, usages, and reservations associated with a
        project.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project being deleted.
        """

        self._driver.destroy_all_by_project(context, project_id)

    def expire(self, context):
        """Expire reservations.

        Explores all currently existing reservations and rolls back
        any that have expired.

        :param context: The request context, for access checks.
        """

        self._driver.expire(context)

    @property
    def resources(self):
        return sorted(self._resources.keys())


QUOTAS = QuotaEngine()


resources = [
    ReservableResource('shares', '_sync_shares', 'shares'),
    ReservableResource('snapshots', '_sync_snapshots', 'snapshots'),
    ReservableResource('gigabytes', '_sync_gigabytes', 'gigabytes'),
    ReservableResource('per_share_gigabytes', None,
                       'per_share_gigabytes'),
    ReservableResource('snapshot_gigabytes', '_sync_snapshot_gigabytes',
                       'snapshot_gigabytes'),
    ReservableResource('share_networks', '_sync_share_networks',
                       'share_networks'),
    ReservableResource('share_groups', '_sync_share_groups',
                       'share_groups'),
    ReservableResource('share_group_snapshots', '_sync_share_group_snapshots',
                       'share_group_snapshots'),
    ReservableResource('share_replicas', '_sync_share_replicas',
                       'share_replicas'),
    ReservableResource('replica_gigabytes', '_sync_replica_gigabytes',
                       'replica_gigabytes'),
]


QUOTAS.register_resources(resources)



# Copyright (c) 2010 OpenStack, LLC.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Simple Scheduler
"""

from oslo_config import cfg

from manila import db
from manila import exception
from manila.i18n import _
from manila.scheduler.drivers import base
from manila.scheduler.drivers import chance
from manila import utils

simple_scheduler_opts = [
    cfg.IntOpt("max_gigabytes",
               default=10000,
               help="Maximum number of volume gigabytes to allow per host."), ]

CONF = cfg.CONF
CONF.register_opts(simple_scheduler_opts)


class SimpleScheduler(chance.ChanceScheduler):
    """Implements Naive Scheduler that tries to find least loaded host."""

    def schedule_create_share(self, context, request_spec, filter_properties):
        """Picks a host that is up and has the fewest shares."""
        # TODO(rushiagr) - pick only hosts that run shares
        elevated = context.elevated()

        share_id = request_spec.get('share_id')
        snapshot_id = request_spec.get('snapshot_id')
        share_properties = request_spec.get('share_properties')
        share_size = share_properties.get('size')

        instance_properties = request_spec.get('share_instance_properties', {})
        availability_zone_id = instance_properties.get('availability_zone_id')

        results = db.service_get_all_share_sorted(elevated)
        if availability_zone_id:
            results = [(service_g, gigs) for (service_g, gigs) in results
                       if (service_g['availability_zone_id']
                           == availability_zone_id)]
        for result in results:
            (service, share_gigabytes) = result
            if share_gigabytes + share_size > CONF.max_gigabytes:
                msg = _("Not enough allocatable share gigabytes remaining")
                raise exception.NoValidHost(reason=msg)
            if utils.service_is_up(service) and not service['disabled']:
                updated_share = base.share_update_db(context,
                                                     share_id,
                                                     service['host'])
                self.share_rpcapi.create_share_instance(
                    context,
                    updated_share.instance,
                    service['host'],
                    request_spec,
                    None,
                    snapshot_id=snapshot_id)
                return None
        msg = _("Is the appropriate service running?")
        raise exception.NoValidHost(reason=msg)



# Copyright (c) 2019 Infortrend Technology, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class InfortrendManilaTestData(object):

    fake_share_id = ['4d6984fd-8572-4467-964f-24936a8c4ea2',  # NFS
                     'a7b933e6-bb77-4823-a86f-f2c3ab41a8a5']  # CIFS

    fake_id = ['iftt8862-2226-0126-7610-chengweichou',
               '987c8763-3333-4444-5555-666666666666']

    fake_share_nfs = {
        'share_id': fake_share_id[0],
        'availability_zone': 'nova',
        'terminated_at': 'datetime.datetime(2017, 5, 8, 8, 27, 25)',
        'availability_zone_id': 'fd32d76d-b5a8-4c5c-93d7-8f09fc2a8ad3',
        'updated_at': 'datetime.datetime(2017, 5, 8, 8, 27, 25)',
        'share_network_id': None,
        'export_locations': [],
        'share_server_id': None,
        'snapshot_id': None,
        'deleted_at': None,
        'id': '5a0aa06e-1c57-4996-be46-b81e360e8866',
        'size': 30,
        'replica_state': None,
        'user_id': '4944594433f0405588928a4212964658',
        'export_location': '172.27.112.223:/share-pool-01/LV-1/' +
        fake_share_id[0],
        'display_description': None,
        'consistency_group_id': None,
        'project_id': '0e63326c50a246ac81fa1a0c8e003d5b',
        'launched_at': 'datetime.datetime(2017, 5, 8, 8, 23, 33)',
        'scheduled_at': 'datetime.datetime(2017, 5, 8, 8, 23, 29)',
        'status': 'deleting',
        'share_type_id': '23d8c637-0192-47fa-b921-958f22ed772f',
        'deleted': 'False',
        'host': 'compute@ift-manila#share-pool-01',
        'access_rules_status': 'active',
        'display_name': 'nfs-01',
        'name': 'share-5a0aa06e-1c57-4996-be46-b81e360e8866',
        'created_at': 'datetime.datetime(2017, 5, 8, 8, 23, 29)',
        'share_proto': 'NFS',
        'is_public': False,
        'source_cgsnapshot_member_id': None
    }

    fake_share_cifs = {
        'share_id': fake_share_id[1],
        'availability_zone': 'nova',
        'terminated_at': None,
        'availability_zone_id': 'fd32d76d-b5a8-4c5c-93d7-8f09fc2a8ad3',
        'updated_at': 'datetime.datetime(2017, 5, 9, 2, 28, 35)',
        'share_network_id': None,
        'export_locations': [],
        'share_server_id': None,
        'snapshot_id': None,
        'deleted_at': None,
        'id': 'aac4fe64-7a9c-472a-b156-9adbb50b4d29',
        'size': 50,
        'replica_state': None,
        'user_id': '4944594433f0405588928a4212964658',
        'export_location': None,
        'display_description': None,
        'consistency_group_id': None,
        'project_id': '0e63326c50a246ac81fa1a0c8e003d5b',
        'launched_at': None,
        'scheduled_at': 'datetime.datetime(2017, 5, 9, 2, 28, 35)',
        'status': 'creating',
        'share_type_id': '23d8c637-0192-47fa-b921-958f22ed772f',
        'deleted': 'False',
        'host': 'compute@ift-manila#share-pool-01',
        'access_rules_status': 'active',
        'display_name': 'cifs-01',
        'name': 'share-aac4fe64-7a9c-472a-b156-9adbb50b4d29',
        'created_at': 'datetime.datetime(2017, 5, 9, 2, 28, 35)',
        'share_proto': 'CIFS',
        'is_public': False,
        'source_cgsnapshot_member_id': None
    }

    fake_share_cifs_no_host = {
        'share_id': fake_share_id[1],
        'availability_zone': 'nova',
        'terminated_at': None,
        'availability_zone_id': 'fd32d76d-b5a8-4c5c-93d7-8f09fc2a8ad3',
        'updated_at': 'datetime.datetime(2017, 5, 9, 2, 28, 35)',
        'share_network_id': None,
        'export_locations': [],
        'share_server_id': None,
        'snapshot_id': None,
        'deleted_at': None,
        'id': 'aac4fe64-7a9c-472a-b156-9adbb50b4d29',
        'size': 50,
        'replica_state': None,
        'user_id': '4944594433f0405588928a4212964658',
        'export_location': None,
        'display_description': None,
        'consistency_group_id': None,
        'project_id': '0e63326c50a246ac81fa1a0c8e003d5b',
        'launched_at': None,
        'scheduled_at': 'datetime.datetime(2017, 5, 9, 2, 28, 35)',
        'status': 'creating',
        'share_type_id': '23d8c637-0192-47fa-b921-958f22ed772f',
        'deleted': 'False',
        'host': '',
        'access_rules_status': 'active',
        'display_name': 'cifs-01',
        'name': 'share-aac4fe64-7a9c-472a-b156-9adbb50b4d29',
        'created_at': 'datetime.datetime(2017, 5, 9, 2, 28, 35)',
        'share_proto': 'CIFS',
        'is_public': False,
        'source_cgsnapshot_member_id': None
    }

    fake_non_exist_share = {
        'share_id': fake_id[0],
        'availability_zone': 'nova',
        'terminated_at': 'datetime.datetime(2017, 5, 8, 8, 27, 25)',
        'availability_zone_id': 'fd32d76d-b5a8-4c5c-93d7-8f09fc2a8ad3',
        'updated_at': 'datetime.datetime(2017, 5, 8, 8, 27, 25)',
        'share_network_id': None,
        'export_locations': [],
        'share_server_id': None,
        'snapshot_id': None,
        'deleted_at': None,
        'id': fake_id[1],
        'size': 30,
        'replica_state': None,
        'user_id': '4944594433f0405588928a4212964658',
        'export_location': '172.27.112.223:/share-pool-01/LV-1/' +
                           fake_id[0],
        'display_description': None,
        'consistency_group_id': None,
        'project_id': '0e63326c50a246ac81fa1a0c8e003d5b',
        'launched_at': 'datetime.datetime(2017, 5, 8, 8, 23, 33)',
        'scheduled_at': 'datetime.datetime(2017, 5, 8, 8, 23, 29)',
        'status': 'available',
        'share_type_id': '23d8c637-0192-47fa-b921-958f22ed772f',
        'deleted': 'False',
        'host': 'compute@ift-manila#share-pool-01',
        'access_rules_status': 'active',
        'display_name': 'nfs-01',
        'name': 'share-5a0aa06e-1c57-4996-be46-b81e360e8866',
        'created_at': 'datetime.datetime(2017, 5, 8, 8, 23, 29)',
        'share_proto': 'NFS',
        'is_public': False,
        'source_cgsnapshot_member_id': None
    }

    fake_access_rules_nfs = [{
        'share_id': fake_share_id[0],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 8, 41, 21)',
        'updated_at': None,
        'access_type': 'ip',
        'access_to': '172.27.1.1',
        'access_level': 'rw',
        'instance_mappings': [],
        'deleted_at': None,
        'id': 'fa60b50f-1428-44a2-9931-7e31f0c5b033'}, {
        'share_id': fake_share_id[0],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 8, 45, 37)',
        'updated_at': None,
        'access_type': 'ip',
        'access_to': '172.27.1.2',
        'access_level': 'rw',
        'instance_mappings': [],
        'deleted_at': None,
        'id': '9bcdd5e6-11c7-4f8f-939c-84fa2f3334bc'
    }]

    fake_rule_ip_1 = [{
        'share_id': fake_share_id[0],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 8, 41, 21)',
        'updated_at': None,
        'access_type': 'ip',
        'access_to': '172.27.1.1',
        'access_level': 'rw',
        'instance_mappings': [],
        'deleted_at': None,
        'id': 'fa60b50f-1428-44a2-9931-7e31f0c5b033'
    }]

    fake_rule_ip_2 = [{
        'share_id': fake_share_id[0],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 8, 45, 37)',
        'updated_at': None,
        'access_type': 'ip',
        'access_to': '172.27.1.2',
        'access_level': 'rw',
        'instance_mappings': [],
        'deleted_at': None,
        'id': '9bcdd5e6-11c7-4f8f-939c-84fa2f3334bc'
    }]

    fake_access_rules_cifs = [{
        'share_id': fake_share_id[1],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 9, 39, 18)',
        'updated_at': None,
        'access_type': 'user',
        'access_to': 'user02',
        'access_level': 'ro',
        'instance_mappings': [],
        'deleted_at': None,
        'id': '6e8bc969-51c9-4bbb-8e8b-020dc5fec81e'}, {
        'share_id': fake_share_id[1],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 9, 38, 59)',
        'updated_at': None,
        'access_type': 'user',
        'access_to': 'user01',
        'access_level': 'rw',
        'instance_mappings': [],
        'deleted_at': None,
        'id': '0cd9926d-fac4-4122-a523-538e98752e78'
    }]

    fake_rule_user01 = [{
        'share_id': fake_share_id[1],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 9, 38, 59)',
        'updated_at': None,
        'access_type': 'user',
        'access_to': 'user01',
        'access_level': 'rw',
        'instance_mappings': [],
        'deleted_at': None,
        'id': '0cd9926d-fac4-4122-a523-538e98752e78'
    }]

    fake_rule_user02 = [{
        'share_id': fake_share_id[1],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 9, 39, 18)',
        'updated_at': None,
        'access_type': 'user',
        'access_to': 'user02',
        'access_level': 'ro',
        'instance_mappings': [],
        'deleted_at': None,
        'id': '6e8bc969-51c9-4bbb-8e8b-020dc5fec81e'
    }]

    fake_rule_user03 = [{
        'share_id': fake_id[0],
        'deleted': 'False',
        'created_at': 'datetime.datetime(2017, 5, 9, 9, 39, 18)',
        'updated_at': None,
        'access_type': 'user',
        'access_to': 'user03',
        'access_level': 'rw',
        'instance_mappings': [],
        'deleted_at': None,
        'id': fake_id[1]
    }]

    fake_share_for_manage_nfs = {
        'share_id': '419ab73c-c0fc-4e73-b56a-70756e0b6d27',
        'availability_zone': None,
        'terminated_at': None,
        'availability_zone_id': None,
        'updated_at': None,
        'share_network_id': None,
        'export_locations': [{
            'uuid': '0ebd59e4-e65e-4fda-9457-320375efd0be',
            'deleted': 0,
            'created_at': 'datetime.datetime(2017, 5, 10, 10, 0, 3)',
            'updated_at': 'datetime.datetime(2017, 5, 10, 10, 0, 3)',
            'is_admin_only': False,
            'share_instance_id': 'd3cfe195-85cf-41e6-be4f-a96f7e7db192',
            'path': '172.27.112.223:/share-pool-01/LV-1/test-folder',
            'el_metadata': {},
            'deleted_at': None,
            'id': 83
        }],
        'share_server_id': None,
        'snapshot_id': None,
        'deleted_at': None,
        'id': '615ac1ed-e808-40b5-8d7b-87018c6f66eb',
        'size': None,
        'replica_state': None,
        'user_id': '4944594433f0405588928a4212964658',
        'export_location': '172.27.112.223:/share-pool-01/LV-1/test-folder',
        'display_description': '',
        'consistency_group_id': None,
        'project_id': '0e63326c50a246ac81fa1a0c8e003d5b',
        'launched_at': None,
        'scheduled_at': 'datetime.datetime(2017, 5, 10, 9, 22, 5)',
        'status': 'manage_starting',
        'share_type_id': '23d8c637-0192-47fa-b921-958f22ed772f',
        'deleted': 'False',
        'host': 'compute@ift-manila#share-pool-01',
        'access_rules_status': 'active',
        'display_name': 'test-manage',
        'name': 'share-615ac1ed-e808-40b5-8d7b-87018c6f66eb',
        'created_at': 'datetime.datetime(2017, 5, 10, 9, 22, 5)',
        'share_proto': 'NFS',
        'is_public': False,
        'source_cgsnapshot_member_id': None
    }

    def _get_fake_share_for_manage(self, location=''):
        return {
            'share_id': '419ab73c-c0fc-4e73-b56a-70756e0b6d27',
            'availability_zone': None,
            'terminated_at': None,
            'availability_zone_id': None,
            'updated_at': None,
            'share_network_id': None,
            'export_locations': [{
                'uuid': '0ebd59e4-e65e-4fda-9457-320375efd0be',
                'deleted': 0,
                'created_at': 'datetime.datetime(2017, 5, 10, 10, 0, 3)',
                'updated_at': 'datetime.datetime(2017, 5, 10, 10, 0, 3)',
                'is_admin_only': False,
                'share_instance_id': 'd3cfe195-85cf-41e6-be4f-a96f7e7db192',
                'path': location,
                'el_metadata': {},
                'deleted_at': None,
                'id': 83
            }],
            'share_server_id': None,
            'snapshot_id': None,
            'deleted_at': None,
            'id': '615ac1ed-e808-40b5-8d7b-87018c6f66eb',
            'size': None,
            'replica_state': None,
            'user_id': '4944594433f0405588928a4212964658',
            'export_location': location,
            'display_description': '',
            'consistency_group_id': None,
            'project_id': '0e63326c50a246ac81fa1a0c8e003d5b',
            'launched_at': None,
            'scheduled_at': 'datetime.datetime(2017, 5, 10, 9, 22, 5)',
            'status': 'manage_starting',
            'share_type_id': '23d8c637-0192-47fa-b921-958f22ed772f',
            'deleted': 'False',
            'host': 'compute@ift-manila#share-pool-01',
            'access_rules_status': 'active',
            'display_name': 'test-manage',
            'name': 'share-615ac1ed-e808-40b5-8d7b-87018c6f66eb',
            'created_at': 'datetime.datetime(2017, 5, 10, 9, 22, 5)',
            'share_proto': 'NFS',
            'is_public': False,
            'source_cgsnapshot_member_id': None
        }

    fake_share_for_manage_cifs = {
        'share_id': '3a1222d3-c981-490a-9390-4d560ced68eb',
        'availability_zone': None,
        'terminated_at': None,
        'availability_zone_id': None,
        'updated_at': None,
        'share_network_id': None,
        'export_locations': [{
            'uuid': '0ebd59e4-e65e-4fda-9457-320375efd0de',
            'deleted': 0,
            'created_at': 'datetime.datetime(2017, 5, 11, 10, 10, 3)',
            'updated_at': 'datetime.datetime(2017, 5, 11, 10, 10, 3)',
            'is_admin_only': False,
            'share_instance_id': 'd3cfe195-85cf-41e6-be4f-a96f7e7db192',
            'path': '\\\\172.27.113.209\\test-folder-02',
            'el_metadata': {},
            'deleted_at': None,
            'id': 87
        }],
        'share_server_id': None,
        'snapshot_id': None,
        'deleted_at': None,
        'id': 'd156baf7-5422-4c9b-8c78-ee7943d000ec',
        'size': None,
        'replica_state': None,
        'user_id': '4944594433f0405588928a4212964658',
        'export_location': '\\\\172.27.113.209\\test-folder-02',
        'display_description': '',
        'consistency_group_id': None,
        'project_id': '0e63326c50a246ac81fa1a0c8e003d5b',
        'launched_at': None,
        'scheduled_at': 'datetime.datetime(2017, 5, 11, 3, 7, 59)',
        'status': 'manage_starting',
        'share_type_id': '23d8c637-0192-47fa-b921-958f22ed772f',
        'deleted': 'False',
        'host': 'compute@ift-manila#share-pool-01',
        'access_rules_status': 'active',
        'display_name': 'test-manage-02',
        'name': 'share-d156baf7-5422-4c9b-8c78-ee7943d000ec',
        'created_at': 'datetime.datetime(2017, 5, 11, 3, 7, 59)',
        'share_proto': 'CIFS',
        'is_public': False,
        'source_cgsnapshot_member_id': None
    }



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Shared File system services driver for Zadara
Virtual Private Storage Array (VPSA).
"""

import socket

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils

from manila import exception as manila_exception
from manila.i18n import _
from manila.share import api
from manila.share import driver
from manila.share.drivers.zadara import common

CONF = cfg.CONF
CONF.register_opts(common.zadara_opts)

LOG = logging.getLogger(__name__)

manila_opts = [
    cfg.StrOpt('zadara_share_name_template',
               default='OS_share-%s',
               help='VPSA - Default template for VPSA share names'),
    cfg.StrOpt('zadara_share_snap_name_template',
               default='OS_share-snapshot-%s',
               help='VPSA - Default template for VPSA share names'),
    cfg.StrOpt('zadara_driver_ssl_cert_path',
               default=None,
               help='Can be used to specify a non default path to a '
                    'CA_BUNDLE file or directory with certificates '
                    'of trusted CAs, which will be used to validate '
                    'the backend')]


class ZadaraVPSAShareDriver(driver.ShareDriver):
    """Zadara VPSA Share driver.

    Version history::

        20.12-01 - Driver changes intended and aligned with
                   openstack latest release.
        20.12-02 - Fixed #18723 - Manila: Parsing the export location in a
                   more generic way while managing the vpsa share
        20.12-03 - Adding the metadata support while creating share to
                   configure vpsa.
        20.12-20 - IPv6 connectivity support for Manila driver
        20.12-21 - Adding unit tests and fixing review comments from the
                   openstack community.
        20.12-22 - Addressing review comments from the manila community.
        20.12-23 - Addressing review comments from the manila community.
        20.12-24 - Addressing review comments from the manila community.
        20.12-25 - Support host assisted share migration
    """

    VERSION = '20.12-25'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "ZadaraStorage_VPSA_CI"

    def __init__(self, *args, **kwargs):
        """Do initialization."""
        super(ZadaraVPSAShareDriver, self).__init__(False, *args, **kwargs)
        self.vpsa = None
        self.configuration.append_config_values(common.zadara_opts)
        self.configuration.append_config_values(manila_opts)
        self.api = api.API()
        # The valid list of share options that can be specified
        # as the metadata while creating manila share
        self.share_options = ['smbguest', 'smbonly', 'smbwindowsacl',
                              'smbfilecreatemask', 'smbbrowseable',
                              'smbhiddenfiles', 'smbhideunreadable',
                              'smbhideunwriteable', 'smbhidedotfiles',
                              'smbstoredosattributes', 'smbdircreatemask',
                              'smbmaparchive', 'smbencryptionmode',
                              'smbenableoplocks', 'smbaiosize',
                              'nfsrootsquash', 'nfsallsquash',
                              'nfsanongid', 'nfsanonuid',
                              'atimeupdate', 'readaheadkb', 'crypt',
                              'compress', 'dedupe', 'attachpolicies']

    def _check_access_key_validity(self):
        try:
            self.vpsa._check_access_key_validity()
        except common.exception.ZadaraInvalidAccessKey:
            raise manila_exception.ZadaraManilaInvalidAccessKey()

    def do_setup(self, context):
        """Any initialization the share driver does while starting.

        Establishes initial connection with VPSA and retrieves access_key.
        Need to pass driver_ssl_cert_path here (and not fetch it from the
        config opts directly in common code), because this config option is
        different for different drivers and so cannot be figured in the
        common code.
        """
        driver_ssl_cert_path = self.configuration.zadara_driver_ssl_cert_path
        self.vpsa = common.ZadaraVPSAConnection(self.configuration,
                                                driver_ssl_cert_path, False)

    def check_for_setup_error(self):
        """Returns an error (exception) if prerequisites aren't met."""
        self._check_access_key_validity()

    def vpsa_send_cmd(self, cmd, **kwargs):
        try:
            response = self.vpsa.send_cmd(cmd, **kwargs)
        except common.exception.UnknownCmd as e:
            raise manila_exception.ZadaraUnknownCmd(cmd=e.cmd)
        except common.exception.SessionRequestException as e:
            raise manila_exception.ZadaraSessionRequestException(msg=e.msg)
        except common.exception.BadHTTPResponseStatus as e:
            raise manila_exception.ZadaraBadHTTPResponseStatus(status=e.status)
        except common.exception.FailedCmdWithDump as e:
            raise manila_exception.ZadaraFailedCmdWithDump(status=e.status,
                                                           data=e.data)
        except common.exception.ZadaraInvalidAccessKey:
            raise manila_exception.ZadaraManilaInvalidAccessKey()
        return response

    def _get_zadara_share_template_name(self, share_id):
        return self.configuration.zadara_share_name_template % share_id

    def _get_share_export_location(self, share):
        export_location = ''
        share_proto = share['share_proto'].upper()

        share_name = self._get_zadara_share_template_name(share['id'])
        vpsa_volume = self.vpsa._get_vpsa_volume(share_name)
        if not vpsa_volume:
            msg = (_('VPSA volume for share %s '
                     'could not be found.') % share['id'])
            LOG.error(msg)
            raise manila_exception.ZadaraShareNotFound(name=share['id'])

        if share_proto == 'NFS':
            export_location = vpsa_volume['nfs_export_path']
        if share_proto == 'CIFS':
            export_location = vpsa_volume['smb_export_path']
        return export_location

    def _check_share_protocol(self, share):
        share_proto = share['share_proto'].upper()
        if share_proto not in ('NFS', 'CIFS'):
            msg = _("Only NFS or CIFS protocol are currently supported. "
                    "Share provided %(share)s with protocol "
                    "%(proto)s.") % {'share': share['id'],
                                     'proto': share['share_proto']}
            LOG.error(msg)
            raise manila_exception.ZadaraInvalidProtocol(
                protocol_type=share_proto)

    def is_valid_metadata(self, metadata):
        LOG.debug('Metadata while creating share: %(metadata)s',
                  {'metadata': metadata})
        for key, value in metadata.items():
            if key in self.share_options:
                # Check for the values allowed with provided metadata
                if key in ['smbguest', 'smbonly', 'smbwindowsacl',
                           'smbbrowseable', 'smbhideunreadable',
                           'smbhideunwriteable', 'smbhidedotfiles',
                           'smbstoredosattributes', 'smbmaparchive',
                           'smbenableoplocks', 'nfsrootsquash',
                           'nfsallsquash', 'atimeupdate', 'crypt',
                           'compress', 'dedupe', 'attachpolicies']:
                    if value in ['YES', 'NO']:
                        continue
                    else:
                        return False
                if key in ['smbfilecreatemask', 'smbdircreatemask']:
                    if value.isdigit():
                        # The valid permissions should be for user,group,other
                        # with another special digit for attributes. Ex:0755
                        if len(value) != 4:
                            return False
                        # No special permission bits for suid,sgid,
                        # stickybit are allowed for vpsa share.
                        if int(value[0]) != 0:
                            return False
                        # The permissions are always specified in octal
                        for i in range(1, len(value)):
                            if int(value[i]) > 7:
                                return False
                        continue
                    else:
                        return False
                if key == 'smbaiosize':
                    if value.isdigit() and value in ['16384', '1']:
                        continue
                    else:
                        return False
                if key == 'smbencryptionmode':
                    if value in ['off', 'desired', 'required']:
                        continue
                    else:
                        return False
                if key in ['nfsanongid', 'nfsanonuid']:
                    if value.isdigit() and int(value) != 0:
                        continue
                    else:
                        return False
                if key == 'readaheadkb':
                    if value in ['16', '64', '128', '256', '512']:
                        continue
                    else:
                        return False
        return True

    def create_share(self, context, share, share_server=None):
        """Create a Zadara share and export it.

        :param context: A RequestContext.
        :param share: A Share.
        :param share_server: Not used currently
        :return: The export locations dictionary.
        """
        # Check share's protocol.
        # Throw an exception immediately if it is an invalid protocol.
        self._check_share_protocol(share)
        share_name = self._get_zadara_share_template_name(share['id'])

        # Collect the share metadata provided and validate it
        metadata = self.api.get_share_metadata(context,
                                               {'id': share['share_id']})
        if not self.is_valid_metadata(metadata):
            raise manila_exception.ManilaException(_(
                "Not a valid metadata provided for the share %s")
                % share['id'])

        data = self.vpsa_send_cmd('create_volume',
                                  name=share_name,
                                  size=share['size'],
                                  metadata=metadata)
        if data['status'] != 0:
            raise manila_exception.ZadaraVPSAVolumeShareFailed(
                error=data['status'])

        export_location = self._get_share_export_location(share)
        return {'path': export_location}

    def _allow_access(self, context, share, access):
        """Allow access to the share."""
        access_type = access['access_type']
        share_proto = share['share_proto'].upper()
        if share_proto == 'CIFS':
            share_proto = 'SMB'

        if access_type != 'ip':
            raise manila_exception.ZadaraInvalidShareAccessType()
        access_ip = access['access_to']
        access_level = 'YES'
        if access['access_level'] == 'rw':
            access_level = 'NO'

        # First: Check Active controller: if not valid, raise exception
        ctrl = self.vpsa._get_active_controller_details()
        if not ctrl:
            raise manila_exception.ZadaraVPSANoActiveController()

        # Get volume name
        vol_name = self._get_zadara_share_template_name(share['id'])
        vpsa_volume = self.vpsa._get_vpsa_volume(vol_name)

        if not vpsa_volume:
            msg = (_('VPSA volume for share %s '
                     'could not be found.') % share['id'])
            LOG.error(msg)
            raise manila_exception.ZadaraShareNotFound(name=share['id'])

        # Get/Create server name for given IP
        vpsa_srv = self.vpsa._create_vpsa_server(iscsi_ip=access_ip)
        if not vpsa_srv:
            raise manila_exception.ZadaraServerCreateFailure(name=access_ip)

        servers = self.vpsa._get_servers_attached_to_volume(vpsa_volume)
        attach = None
        for server in servers:
            if server == vpsa_srv:
                attach = server
                break
        # Attach volume to server
        if attach is None:
            self.vpsa_send_cmd('attach_volume',
                               vpsa_srv=vpsa_srv,
                               vpsa_vol=vpsa_volume['name'],
                               share_proto=share_proto,
                               read_only=access_level)

        data = self.vpsa_send_cmd('list_vol_attachments',
                                  vpsa_vol=vpsa_volume['name'])
        server = None
        servers = data.get('servers', [])
        for srv in servers:
            if srv['iscsi_ip'] == access_ip:
                server = srv
                break

        if server is None:
            raise manila_exception.ZadaraAttachmentsNotFound(
                name=vpsa_volume['name'])

        ctrl_ip = self.vpsa._get_target_host(ctrl['ip'])
        properties = {'target_discovered': False,
                      'target_portal': (('%s:%s') % (ctrl_ip, '3260')),
                      'target_ip': server['iscsi_ip'],
                      'id': share['id'],
                      'auth_method': 'CHAP',
                      'auth_username': ctrl['chap_user'],
                      'auth_password': ctrl['chap_passwd']}

        LOG.debug('Attach properties: %(properties)s',
                  {'properties': strutils.mask_password(properties)})
        return {'driver_volume_type': share['share_proto'], 'data': properties}

    def delete_share(self, context, share, share_server=None):
        """Delete share. Auto detach from all servers.

        """
        # Get share name
        share_name = self._get_zadara_share_template_name(share['id'])
        volume = self.vpsa._get_vpsa_volume(share_name)
        if not volume:
            LOG.warning('Volume %s could not be found. '
                        'It might be already deleted', share['id'])
            return

        self.vpsa._detach_vpsa_volume(vpsa_vol=volume)

        # Delete volume associate with the share
        self.vpsa_send_cmd('delete_volume', vpsa_vol=volume['name'])

    def _deny_access(self, context, share, access, share_server=None):
        """Deny access to the share from the host.

        """
        access_type = access['access_type']
        if access_type != 'ip':
            LOG.warning('Only ip access type is allowed for zadara vpsa.')
            return
        access_ip = access['access_to']

        # First: Check Active controller: if not valid, raise exception
        ctrl = self.vpsa._get_active_controller_details()
        if not ctrl:
            raise manila_exception.ZadaraVPSANoActiveController()

        # Get share name
        share_name = self._get_zadara_share_template_name(share['id'])
        volume = self.vpsa._get_vpsa_volume(share_name)
        if not volume:
            LOG.warning('Volume %s could not be found. '
                        'It might be already deleted', share['id'])
            return

        vpsa_srv = self.vpsa._get_server_name(access_ip, True)
        if not vpsa_srv:
            LOG.warning('VPSA server %s could not be found.', access_ip)
            return

        servers_list = self.vpsa._get_servers_attached_to_volume(volume)
        if vpsa_srv not in servers_list:
            LOG.warning('VPSA server %(access_ip)s not attached '
                        'to volume %(volume)s.',
                        {'access_ip': access_ip, 'volume': share['id']})
            return

        self.vpsa._detach_vpsa_volume(vpsa_vol=volume,
                                      vpsa_srv=vpsa_srv)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        access_updates = {}
        if not (add_rules or delete_rules):
            # add_rules and delete_rules can be empty lists, in cases
            # like share migration for zadara driver, when the access
            # level is to be changed for all existing rules. For zadara
            # backend, we delete and re-add all the existing rules.
            for access_rule in access_rules:
                self._deny_access(context, share, access_rule)
                try:
                    self._allow_access(context, share, access_rule)
                except manila_exception.ZadaraInvalidShareAccessType:
                    LOG.error("Only ip access type allowed for Zadara share. "
                              "Failed to allow %(access_level)s access to "
                              "%(access_to)s for rule %(id)s. Setting rule "
                              "to 'error' state.",
                              {'access_level': access_rule['access_level'],
                               'access_to': access_rule['access_to'],
                               'id': access_rule['access_id']})
                    access_updates.update(
                        {access_rule['access_id']: {'state': 'error'}})
        else:
            if add_rules:
                # Add rules for accessing share
                for access_rule in add_rules:
                    try:
                        self._allow_access(context, share, access_rule)
                    except manila_exception.ZadaraInvalidShareAccessType:
                        LOG.error("Only ip access type allowed for Zadara "
                                  "share. Failed to allow %(access_level)s "
                                  "access to %(access_to)s for rule %(id)s. "
                                  "Setting rule to 'error' state.",
                                  {'access_level': access_rule['access_level'],
                                   'access_to': access_rule['access_to'],
                                   'id': access_rule['access_id']})
                        access_updates.update(
                            {access_rule['access_id']: {'state': 'error'}})
            if delete_rules:
                # Delete access rules for provided share
                for access_rule in delete_rules:
                    self._deny_access(context, share, access_rule)
        return access_updates

    def extend_share(self, share, new_size, share_server=None):
        """Extend an existing share.

        """
        # Get the backend volume name for the share
        share_name = self._get_zadara_share_template_name(share['id'])
        vpsa_volume = self.vpsa._get_vpsa_volume(share_name)
        if not vpsa_volume:
            msg = (_('VPSA volume for share %s '
                     'could not be found.') % share['id'])
            LOG.error(msg)
            raise manila_exception.ZadaraShareNotFound(name=share['id'])

        size = vpsa_volume['virtual_capacity']
        expand_size = new_size - size
        data = self.vpsa_send_cmd('expand_volume',
                                  vpsa_vol=vpsa_volume['name'],
                                  size=expand_size)
        if data['status'] != 0:
            raise manila_exception.ZadaraExtendShareFailed(
                error=data['status'])

    def _ensure_share(self, context, share, share_server=None):
        """Ensure that the share has a backend volume and it is exported.

        """
        # Get the backend volume name for the share
        share_name = self._get_zadara_share_template_name(share['id'])
        vpsa_volume = self.vpsa._get_vpsa_volume(share_name)
        if not vpsa_volume:
            msg = (_('VPSA volume for share %s '
                     'could not be found.') % share['id'])
            LOG.error(msg)
            raise manila_exception.ZadaraShareNotFound(name=share['id'])

        export_locations = share['export_locations']
        if export_locations:
            return export_locations
        else:
            servers_list = (self.vpsa._get_servers_attached_to_volume(
                            vpsa_volume))
            if len(servers_list) != 0:
                msg = (_('Servers attached to the VPSA volume %s without '
                         'any locations exported.') % vpsa_volume['name'])
                LOG.error(msg)
                raise manila_exception.ZadaraShareNotValid(
                    name=share['id'])

    def _update_share_stats(self):

        backend_name = self.configuration.share_backend_name
        dhss = self.configuration.driver_handles_share_servers
        vpsa_poolname = self.configuration.zadara_vpsa_poolname
        (total, free, provisioned) = (
            self.vpsa._get_pool_capacity(vpsa_poolname))
        ctrl = self.vpsa._get_active_controller_details()
        if not ctrl:
            raise manila_exception.ZadaraVPSANoActiveController()
        ipv4_support = False if ':' in ctrl['ip'] else True

        # VPSA backend pool
        single_pool = dict(
            pool_name=vpsa_poolname,
            total_capacity_gb=total,
            free_capacity_gb=free,
            allocated_capacity_gb=(total - free),
            provisioned_capacity_gb=provisioned,
            reserved_percentage=self.configuration.reserved_share_percentage,
            reserved_snapshot_percentage=(
                self.configuration.reserved_share_from_snapshot_percentage
                or self.configuration.reserved_share_percentage),
            compression=[True, False],
            dedupe=[True, False],
            thin_provisioning=True
        )

        data = dict(
            share_backend_name=backend_name,
            driver_handles_share_servers=dhss,
            vendor_name='Zadara Storage',
            driver_version=self.VERSION,
            storage_protocol='NFS_CIFS',
            pools=[single_pool],
            snapshot_support=True,
            create_share_from_snapshot_support=True,
            revert_to_snapshot_support=False,
            mount_snapshot_support=False,
            ipv4_support=ipv4_support,
            ipv6_support=not ipv4_support
        )
        super(ZadaraVPSAShareDriver, self)._update_share_stats(data)

    def create_snapshot(self, context, snapshot, share_server=None):
        """Creates a snapshot."""
        LOG.debug('Create snapshot: %s', snapshot['id'])

        # Retrieve the CG name for the base volume
        share = snapshot['share']
        volume_name = self._get_zadara_share_template_name(share['id'])
        cg_name = self.vpsa._get_volume_cg_name(volume_name)
        if not cg_name:
            msg = (_('VPSA volume for share %s '
                     'could not be found.') % share['id'])
            LOG.error(msg)
            raise manila_exception.ZadaraShareNotFound(name=share['id'])

        snap_name = (self.configuration.zadara_share_snap_name_template
                     % snapshot['id'])
        data = self.vpsa_send_cmd('create_snapshot',
                                  cg_name=cg_name,
                                  snap_name=snap_name)
        if data['status'] != 0:
            raise manila_exception.ZadaraVPSASnapshotCreateFailed(
                name=share['id'], error=data['status'])

        return {'provider_location': data['snapshot_name']}

    def delete_snapshot(self, context, snapshot, share_server=None):
        """Deletes a snapshot."""
        LOG.debug('Delete snapshot: %s', snapshot['id'])

        # Retrieve the CG name for the base volume
        share = snapshot['share']
        volume_name = self._get_zadara_share_template_name(share['id'])
        cg_name = self.vpsa._get_volume_cg_name(volume_name)
        if not cg_name:
            # If the volume isn't present, then don't attempt to delete
            LOG.warning('snapshot: original volume %s not found, '
                        'skipping delete operation',
                        volume_name)
            return

        snap_name = (self.configuration.zadara_share_snap_name_template
                     % snapshot['id'])
        snap_id = self.vpsa._get_snap_id(cg_name, snap_name)
        if not snap_id:
            # If the snapshot isn't present, then don't attempt to delete
            LOG.warning('snapshot: snapshot %s not found, '
                        'skipping delete operation', snap_name)
            return

        self.vpsa_send_cmd('delete_snapshot',
                           snap_id=snap_id)

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Creates a share from a snapshot.

        """
        LOG.debug('Creating share from snapshot: %s', snapshot['id'])

        # Retrieve the CG name for the base volume
        volume_name = (self._get_zadara_share_template_name(
                       snapshot['share_instance_id']))
        cg_name = self.vpsa._get_volume_cg_name(volume_name)
        if not cg_name:
            msg = (_('VPSA volume for share %s '
                     'could not be found.') % share['id'])
            LOG.error(msg)
            raise manila_exception.ZadaraShareNotFound(name=share['id'])

        snap_name = (self.configuration.zadara_share_snap_name_template
                     % snapshot['id'])
        snap_id = self.vpsa._get_snap_id(cg_name, snap_name)
        if not snap_id:
            msg = _('Snapshot %(name)s not found') % {'name': snap_name}
            LOG.error(msg)
            raise manila_exception.ShareSnapshotNotFound(
                snapshot_id=snap_name)

        self._check_share_protocol(share)

        share_name = self._get_zadara_share_template_name(share['id'])
        self.vpsa_send_cmd('create_clone_from_snap',
                           cg_name=cg_name,
                           name=share_name,
                           snap_id=snap_id)

        if share['size'] > snapshot['size']:
            self.extend_share(share, share['size'])

        export_location = self._get_share_export_location(share)
        return [{'path': export_location}]

    def _get_export_name_from_export_path(self, proto, export_path):
        if proto == 'nfs' and '\\' in export_path:
            return None
        if proto == 'cifs' and '/' in export_path:
            return None

        # Extract the export name from the provided export path
        if proto == 'nfs':
            separator = '/'
            export_location = export_path.strip(separator)
            export_name = export_location.split(separator)[-1]
        else:
            separator = '\\'
            export_location = export_path.strip(separator)
            export_name = export_location.split(separator)[-1]
        return export_name

    def _extract_vpsa_volume_from_share(self, share):
        """Returns a vpsa volume based on the export location"""
        if not share['export_locations'][0]['path']:
            return None

        share_proto = share['share_proto'].lower()
        export_path = share['export_locations'][0]['path']
        export_name = self._get_export_name_from_export_path(share_proto,
                                                             export_path)
        if export_name is None:
            msg = (_('Please verify the specifed protocol and export path.'))
            LOG.error(msg)
            raise manila_exception.ManilaException(msg)

        volume = None
        volumes = self.vpsa._get_all_vpsa_volumes()
        # Find the volume with the corresponding export name
        for vol in volumes:
            if share_proto == 'nfs':
                vol_export_path = vol.get('nfs_export_path', None)
            else:
                vol_export_path = vol.get('smb_export_path', None)

            vol_export_name = self._get_export_name_from_export_path(
                share_proto, vol_export_path)
            if export_name == vol_export_name:
                volume = vol
                break

            # Check the additional smb export paths of the volume
            if (share_proto == 'cifs' and
                    vol['additional_smb_export_paths_count'] > 0):
                for additional_path in vol['additional_smb_export_paths']:
                    vol_export_name = self._get_export_name_from_export_path(
                        share_proto, additional_path)
                    if export_name == vol_export_name:
                        volume = vol
                        break
        if volume:
            return volume
        else:
            msg = (_('Manage backend share could not be found. It might be '
                     'deleted or please verify the specifed protocol and '
                     'export path.'))
            LOG.error(msg)
            raise manila_exception.ManilaException(msg)

    def manage_existing(self, share, driver_options):
        # Check whether the specified protocol is supported or not.
        self._check_share_protocol(share)

        LOG.info("Share %(shr_path)s will be managed with share %(shr_name)s.",
                 {'shr_path': share['export_locations'][0]['path'],
                  'shr_name': share['id']})

        # Find the backend vpsa volume for the provided export location
        vpsa_volume = self._extract_vpsa_volume_from_share(share)

        # Check if the volume is available
        if vpsa_volume['status'] != 'Available':
            msg = (_('Existing share %(name)s is not available')
                   % {'name': vpsa_volume['name']})
            LOG.error(msg)
            raise manila_exception.ManilaException(msg)

        new_share_name = self._get_zadara_share_template_name(share['id'])
        new_vpsa_share = self.vpsa._get_vpsa_volume(new_share_name)
        if new_vpsa_share:
            msg = (_('Share %(new_name)s already exists')
                   % {'new_name': new_share_name})
            LOG.error(msg)
            raise manila_exception.ManilaException(msg)

        # Rename the volume to the manila share specified name
        data = self.vpsa_send_cmd('rename_volume',
                                  vpsa_vol=vpsa_volume['name'],
                                  new_name=new_share_name)
        if data['status'] != 0:
            msg = (_('Renaming volume %(old_name)s to %(new_name)s '
                     'has failed.') % {'old_name': vpsa_volume['name'],
                                       'new_name': new_share_name})
            LOG.error(msg)
            raise manila_exception.ManilaException(msg)

        return {'size': vpsa_volume['provisioned_capacity'],
                'export_locations': share['export_locations'][0]['path']}

    def unmanage(self, share):
        """Removes the specified volume from Manila management"""
        pass

    def manage_existing_snapshot(self, snapshot, driver_options):
        share = snapshot['share']
        share_name = self._get_zadara_share_template_name(share['id'])

        vpsa_volume = self.vpsa._get_vpsa_volume(share_name)
        if not vpsa_volume:
            msg = (_('Volume %(name)s could not be found. '
                     'It might be already deleted') % {'name': share_name})
            LOG.error(msg)
            raise manila_exception.ZadaraShareNotFound(name=share['id'])

        # Check if the provider_location is specified
        if not snapshot['provider_location']:
            msg = (_('Provider location as snap id of the VPSA backend '
                     'should be provided'))
            LOG.error(msg)
            raise manila_exception.ManilaException(msg)

        new_name = (self.configuration.zadara_share_snap_name_template
                    % snapshot['id'])
        new_snap_id = self.vpsa._get_snap_id(vpsa_volume['cg_name'],
                                             new_name)
        if new_snap_id:
            msg = (_('Snapshot with name %s already exists') % new_name)
            LOG.debug(msg)
            return

        data = self.vpsa_send_cmd('rename_snapshot',
                                  snap_id=snapshot['provider_location'],
                                  new_name=new_name)
        if data['status'] != 0:
            raise manila_exception.ZadaraVPSASnapshotManageFailed(
                snap_id=snapshot['provider_location'],
                error=data['status'])

    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Manila management"""
        pass

    def get_configured_ip_versions(self):
        """"Get allowed IP versions.

        The shares created should have export location as per the
        IP version. Currently, zadara backend doesn't support both
        ipv4 and ipv6. Collect the supported IP version from the
        vpsa's active controller
        """
        ctrl = self.vpsa._get_active_controller_details()
        if not ctrl:
            raise manila_exception.ZadaraVPSANoActiveController()

        if ':' in ctrl['ip']:
            return [6]
        else:
            return [4]

    def get_backend_info(self, context):
        return {
            'version': self.VERSION,
            'vsa_feip': socket.gethostbyname(self.vpsa.conf.zadara_vpsa_host),
            'vsa_port': self.vpsa.conf.zadara_vpsa_port
        }

    def ensure_shares(self, context, shares):
        updates = {}
        for share in shares:
            updates[share['id']] = {
                'export_locations': self._ensure_share(context, share)}
        return updates



# Copyright 2016 SAP SE
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""change_user_id_length

Revision ID: 221a83cfd85b
Revises: eb6d5544cbbd
Create Date: 2016-06-21 14:22:48.314501

"""

# revision identifiers, used by Alembic.
revision = '221a83cfd85b'
down_revision = 'eb6d5544cbbd'

from alembic import op
from oslo_log import log
import sqlalchemy as sa

LOG = log.getLogger(__name__)


def upgrade():
    LOG.info("Changing user_id length for share_networks")
    op.alter_column("share_networks", "user_id",
                    type_=sa.String(length=255))

    LOG.info("Changing project_id length for share_networks")
    op.alter_column("share_networks", "project_id",
                    type_=sa.String(length=255))

    LOG.info("Changing project_id length for security_services")
    op.alter_column("security_services", "project_id",
                    type_=sa.String(length=255))


def downgrade():
    LOG.info("Changing back user_id length for share_networks")
    op.alter_column("share_networks", "user_id",
                    type_=sa.String(length=36))

    LOG.info("Changing back project_id length for share_networks")
    op.alter_column("share_networks", "project_id",
                    type_=sa.String(length=36))

    LOG.info("Changing back project_id length for security_services")
    op.alter_column("security_services", "project_id",
                    type_=sa.String(length=36))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add 'provider_location' attr to 'share_group_snapshot_members' model.

Revision ID: 927920b37453
Revises: a77e2ad5012d
Create Date: 2017-01-31 20:10:44.937763

"""

# revision identifiers, used by Alembic.
revision = '927920b37453'
down_revision = 'a77e2ad5012d'

from alembic import op
import sqlalchemy as sa


SGSM_TABLE_NAME = 'share_group_snapshot_members'
PROVIDER_LOCATION_NAME = 'provider_location'


def upgrade():
    op.add_column(
        SGSM_TABLE_NAME,
        sa.Column(PROVIDER_LOCATION_NAME, sa.String(255), nullable=True),
    )


def downgrade():
    op.drop_column(SGSM_TABLE_NAME, PROVIDER_LOCATION_NAME)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from webob import exc

from manila.api.openstack import api_version_request as api_version
from manila.api.v1 import share_servers
from manila.common import constants
from manila import context
from manila.db import api as db_api
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes

fake_share_server_list = {
    'share_servers': [
        {
            'status': constants.STATUS_ACTIVE,
            'updated_at': None,
            'host': 'fake_host',
            'share_network_name': 'fake_sn_name',
            'share_network_id': 'fake_sn_id',
            'share_network_subnet_id': 'fake_sns_id',
            'project_id': 'fake_project_id',
            'id': 'fake_server_id',
            'is_auto_deletable': False,
            'task_state': None,
            'source_share_server_id': None,
            'identifier': 'fake_id',
            'security_service_update_support': False
        },
        {
            'status': constants.STATUS_ERROR,
            'updated_at': None,
            'host': 'fake_host_2',
            'share_network_name': 'fake_sn_id_2',
            'share_network_id': 'fake_sn_id_2',
            'share_network_subnet_id': 'fake_sns_id_2',
            'project_id': 'fake_project_id_2',
            'id': 'fake_server_id_2',
            'is_auto_deletable': True,
            'task_state': None,
            'source_share_server_id': None,
            'identifier': 'fake_id_2',
            'security_service_update_support': False

        },
    ]
}

fake_share_network_get_list = {
    'share_networks': [
        {
            'name': 'fake_sn_name',
            'id': 'fake_sn_id',
            'project_id': 'fake_project_id',
        },
        {
            'name': None,
            'id': 'fake_sn_id_2',
            'project_id': 'fake_project_id_2',
        }
    ]
}

fake_share_server_get_result = {
    'share_server': {
        'status': constants.STATUS_ACTIVE,
        'created_at': None,
        'updated_at': None,
        'host': 'fake_host',
        'share_network_name': 'fake_sn_name',
        'share_network_id': 'fake_sn_id',
        'share_network_subnet_id': 'fake_sns_id',
        'project_id': 'fake_project_id',
        'id': 'fake_server_id',
        'backend_details': {
            'fake_key_1': 'fake_value_1',
            'fake_key_2': 'fake_value_2',
        },
        'is_auto_deletable': False,
        'task_state': None,
        'source_share_server_id': None,
        'identifier': 'fake_id',
        'security_service_update_support': False
    }
}

share_server_backend_details = {
    'fake_key_1': 'fake_value_1',
    'fake_key_2': 'fake_value_2',
}

fake_share_server_backend_details_get_result = {
    'details': share_server_backend_details
}


CONTEXT = context.get_admin_context()


class FakeShareServer(object):

    def __init__(self, *args, **kwargs):
        super(FakeShareServer, self).__init__()
        self.id = kwargs.get('id', 'fake_server_id')
        if 'created_at' in kwargs:
            self.created_at = kwargs.get('created_at', None)
        self.updated_at = kwargs.get('updated_at', None)
        self.host = kwargs.get('host', 'fake_host')
        self.share_network_subnet = kwargs.get('share_network_subnet', {
            'id': 'fake_sns_id', 'share_network_id': 'fake_sn_id'})
        self.share_network_subnet_id = kwargs.get(
            'share_network_subnet_id', self.share_network_subnet['id'])
        self.status = kwargs.get('status', constants.STATUS_ACTIVE)
        self.project_id = 'fake_project_id'
        self.identifier = kwargs.get('identifier', 'fake_id')
        self.is_auto_deletable = kwargs.get('is_auto_deletable', False)
        self.task_state = kwargs.get('task_state')
        self.source_share_server_id = kwargs.get('source_share_server_id')
        self.backend_details = share_server_backend_details
        self.security_service_update_support = kwargs.get(
            'security_service_update_support', False)

    def __getitem__(self, item):
        return getattr(self, item)


def fake_share_server_get_all():
    fake_share_servers = [
        FakeShareServer(),
        FakeShareServer(id='fake_server_id_2',
                        host='fake_host_2',
                        share_network_subnet={
                            'id': 'fake_sns_id_2',
                            'share_network_id': 'fake_sn_id_2',
                            },
                        identifier='fake_id_2',
                        task_state=None,
                        is_auto_deletable=True,
                        status=constants.STATUS_ERROR,
                        security_service_update_support=False)
    ]
    return fake_share_servers


def fake_share_server_get():
    return FakeShareServer(created_at=None)


class FakeRequestAdmin(object):
    environ = {"manila.context": CONTEXT}
    GET = {}


class FakeRequestWithHost(FakeRequestAdmin):
    GET = {'host': fake_share_server_list['share_servers'][0]['host']}


class FakeRequestWithStatus(FakeRequestAdmin):
    GET = {'status': constants.STATUS_ERROR}


class FakeRequestWithProjectId(FakeRequestAdmin):
    GET = {'project_id': fake_share_server_get_all()[0].project_id}


class FakeRequestWithShareNetworkSubnetId(FakeRequestAdmin):
    GET = {
        'share_network_subnet_id':
            fake_share_server_get_all()[0].share_network_subnet_id,
    }


class FakeRequestWithFakeFilter(FakeRequestAdmin):
    GET = {'fake_key': 'fake_value'}


@ddt.ddt
class ShareServerAPITest(test.TestCase):

    def setUp(self):
        super(ShareServerAPITest, self).setUp()
        self.controller = share_servers.ShareServerController()
        self.resource_name = self.controller.resource_name
        self.mock_object(policy, 'check_policy',
                         mock.Mock(return_value=True))
        self.mock_object(db_api, 'share_server_get_all',
                         mock.Mock(return_value=fake_share_server_get_all()))

    def _prepare_request(self, url, use_admin_context,
                         version=api_version._MAX_API_VERSION):
        request = fakes.HTTPRequest.blank(url,
                                          use_admin_context=use_admin_context,
                                          version=version)
        ctxt = request.environ['manila.context']
        return request, ctxt

    def test_index_no_filters(self):
        request, ctxt = self._prepare_request(url='/v2/share-servers/',
                                              use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            side_effect=[fake_share_network_get_list['share_networks'][0],
                         fake_share_network_get_list['share_networks'][1]]))
        result = self.controller.index(request)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        db_api.share_server_get_all.assert_called_once_with(ctxt)
        self.assertEqual(fake_share_server_list, result)

    def test_index_host_filter(self):
        request, ctxt = self._prepare_request(
            url='/index?host=%s'
                % fake_share_server_list['share_servers'][0]['host'],
            use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            side_effect=[fake_share_network_get_list['share_networks'][0],
                         fake_share_network_get_list['share_networks'][1]]))
        result = self.controller.index(request)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        db_api.share_server_get_all.assert_called_once_with(ctxt)
        self.assertEqual([fake_share_server_list['share_servers'][0]],
                         result['share_servers'])

    def test_index_status_filter(self):
        request, ctxt = self._prepare_request(url='/index?status=%s' %
                                                  constants.STATUS_ERROR,
                                              use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            side_effect=[fake_share_network_get_list['share_networks'][0],
                         fake_share_network_get_list['share_networks'][1]]))
        result = self.controller.index(request)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        db_api.share_server_get_all.assert_called_once_with(ctxt)
        self.assertEqual([fake_share_server_list['share_servers'][1]],
                         result['share_servers'])

    def test_index_project_id_filter(self):
        request, ctxt = self._prepare_request(
            url='/index?project_id=%s'
                % fake_share_server_get_all()[0].project_id,
            use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            side_effect=[fake_share_network_get_list['share_networks'][0],
                         fake_share_network_get_list['share_networks'][1]]))
        result = self.controller.index(request)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        db_api.share_server_get_all.assert_called_once_with(ctxt)

        self.assertEqual([fake_share_server_list['share_servers'][0]],
                         result['share_servers'])

    def test_index_share_network_filter_by_name(self):
        request, ctxt = self._prepare_request(
            url='/index?host=%s'
                % fake_share_server_list['share_servers'][0]['host'],
            use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            side_effect=[fake_share_network_get_list['share_networks'][0],
                         fake_share_network_get_list['share_networks'][1]]))
        result = self.controller.index(request)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        db_api.share_server_get_all.assert_called_once_with(ctxt)
        self.assertEqual([fake_share_server_list['share_servers'][0]],
                         result['share_servers'])

    def test_index_share_network_filter_by_id(self):
        request, ctxt = self._prepare_request(
            url='/index?share_network=%s'
                % fake_share_network_get_list['share_networks'][0]['id'],
            use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            side_effect=[fake_share_network_get_list['share_networks'][0],
                         fake_share_network_get_list['share_networks'][1]]))
        result = self.controller.index(request)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        db_api.share_server_get_all.assert_called_once_with(ctxt)
        self.assertEqual([fake_share_server_list['share_servers'][0]],
                         result['share_servers'])

    def test_index_fake_filter(self):
        request, ctxt = self._prepare_request(url='/index?fake_key=fake_value',
                                              use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            side_effect=[fake_share_network_get_list['share_networks'][0],
                         fake_share_network_get_list['share_networks'][1]]))
        result = self.controller.index(request)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        db_api.share_server_get_all.assert_called_once_with(ctxt)
        self.assertEqual(0, len(result['share_servers']))

    def test_index_share_network_not_found(self):
        request, ctxt = self._prepare_request(
            url='/index?identifier=%s'
                % fake_share_server_get_all()[0].identifier,
            use_admin_context=True)
        self.mock_object(
            db_api, 'share_network_get',
            mock.Mock(side_effect=exception.ShareNetworkNotFound(
                share_network_id='fake')))

        result = self.controller.index(request)
        db_api.share_server_get_all.assert_called_once_with(ctxt)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        exp_share_server = fake_share_server_list['share_servers'][0].copy()
        exp_share_server['project_id'] = ''
        exp_share_server['share_network_name'] = ''
        self.assertEqual([exp_share_server],
                         result['share_servers'])

    def test_index_share_network_not_found_filter_project(self):
        request, ctxt = self._prepare_request(
            url='/index?project_id=%s'
                % fake_share_server_get_all()[0].project_id,
            use_admin_context=True)
        self.mock_object(
            db_api, 'share_network_get',
            mock.Mock(side_effect=exception.ShareNetworkNotFound(
                share_network_id='fake')))

        result = self.controller.index(request)
        db_api.share_server_get_all.assert_called_once_with(ctxt)
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'index')
        self.assertEqual(0, len(result['share_servers']))

    def test_show(self):
        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(return_value=fake_share_server_get()))
        request, ctxt = self._prepare_request('/show', use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            return_value=fake_share_network_get_list['share_networks'][0]))
        result = self.controller.show(
            request,
            fake_share_server_get_result['share_server']['id'])
        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'show')
        db_api.share_server_get.assert_called_once_with(
            ctxt, fake_share_server_get_result['share_server']['id'])
        self.assertEqual(fake_share_server_get_result['share_server'],
                         result['share_server'])

    @ddt.data(
        {'share_server_side_effect': exception.ShareServerNotFound(
            share_server_id="foo"),
            'share_net_side_effect': mock.Mock()},
        {'share_server_side_effect': mock.Mock(
            return_value=fake_share_server_get()),
            'share_net_side_effect': exception.ShareNetworkNotFound(
                share_network_id="foo")})
    @ddt.unpack
    def test_show_server_not_found(self, share_server_side_effect,
                                   share_net_side_effect):
        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(side_effect=share_server_side_effect))
        request, ctxt = self._prepare_request('/show', use_admin_context=True)
        self.mock_object(db_api, 'share_network_get', mock.Mock(
            side_effect=share_net_side_effect))
        self.assertRaises(
            exc.HTTPNotFound, self.controller.show, request,
            fake_share_server_get_result['share_server']['id'])

        policy.check_policy.assert_called_once_with(
            ctxt, self.resource_name, 'show')
        db_api.share_server_get.assert_called_once_with(
            ctxt, fake_share_server_get_result['share_server']['id'])
        if isinstance(share_net_side_effect, exception.ShareNetworkNotFound):
            exp_share_net_id = (fake_share_server_get()
                                .share_network_subnet['share_network_id'])
            db_api.share_network_get.assert_called_once_with(
                ctxt, exp_share_net_id)

    def test_details(self):
        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(return_value=fake_share_server_get()))
        result = self.controller.details(
            FakeRequestAdmin,

            fake_share_server_get_result['share_server']['id'])
        policy.check_policy.assert_called_once_with(
            CONTEXT, self.resource_name, 'details')
        db_api.share_server_get.assert_called_once_with(
            CONTEXT, fake_share_server_get_result['share_server']['id'])
        self.assertEqual(fake_share_server_backend_details_get_result,
                         result)

    def test_details_share_server_not_found(self):
        share_server_id = 'fake'
        self.mock_object(
            db_api, 'share_server_get',
            mock.Mock(side_effect=exception.ShareServerNotFound(
                share_server_id=share_server_id)))
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.details,
                          FakeRequestAdmin,
                          share_server_id)
        policy.check_policy.assert_called_once_with(
            CONTEXT, self.resource_name, 'details')
        db_api.share_server_get.assert_called_once_with(
            CONTEXT, share_server_id)

    def test_delete_active_server(self):
        share_server = FakeShareServer(status=constants.STATUS_ACTIVE)
        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(return_value=share_server))
        self.mock_object(self.controller.share_api, 'delete_share_server')
        self.controller.delete(
            FakeRequestAdmin,
            fake_share_server_get_result['share_server']['id'])
        policy.check_policy.assert_called_once_with(
            CONTEXT, self.resource_name, 'delete')
        db_api.share_server_get.assert_called_once_with(
            CONTEXT, fake_share_server_get_result['share_server']['id'])
        self.controller.share_api.delete_share_server.assert_called_once_with(
            CONTEXT, share_server)

    def test_delete_error_server(self):
        share_server = FakeShareServer(status=constants.STATUS_ERROR)
        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(return_value=share_server))
        self.mock_object(self.controller.share_api, 'delete_share_server')
        self.controller.delete(
            FakeRequestAdmin,
            fake_share_server_get_result['share_server']['id'])
        policy.check_policy.assert_called_once_with(
            CONTEXT, self.resource_name, 'delete')
        db_api.share_server_get.assert_called_once_with(
            CONTEXT, fake_share_server_get_result['share_server']['id'])
        self.controller.share_api.delete_share_server.assert_called_once_with(
            CONTEXT, share_server)

    def test_delete_used_server(self):
        share_server_id = fake_share_server_get_result['share_server']['id']

        def raise_not_share_server_in_use(*args, **kwargs):
            raise exception.ShareServerInUse(share_server_id=share_server_id)

        share_server = fake_share_server_get()
        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(return_value=share_server))
        self.mock_object(self.controller.share_api, 'delete_share_server',
                         mock.Mock(side_effect=raise_not_share_server_in_use))
        self.assertRaises(exc.HTTPConflict,
                          self.controller.delete,
                          FakeRequestAdmin,
                          share_server_id)
        db_api.share_server_get.assert_called_once_with(CONTEXT,
                                                        share_server_id)
        self.controller.share_api.delete_share_server.assert_called_once_with(
            CONTEXT, share_server)

    def test_delete_not_found(self):
        share_server_id = fake_share_server_get_result['share_server']['id']

        def raise_not_found(*args, **kwargs):
            raise exception.ShareServerNotFound(
                share_server_id=share_server_id)

        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(side_effect=raise_not_found))
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.delete,
                          FakeRequestAdmin,
                          share_server_id)
        db_api.share_server_get.assert_called_once_with(
            CONTEXT, share_server_id)
        policy.check_policy.assert_called_once_with(
            CONTEXT, self.resource_name, 'delete')

    def test_delete_creating_server(self):
        share_server = FakeShareServer(status=constants.STATUS_CREATING)
        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(return_value=share_server))
        self.assertRaises(exc.HTTPForbidden,
                          self.controller.delete,
                          FakeRequestAdmin,
                          share_server['id'])
        policy.check_policy.assert_called_once_with(
            CONTEXT,
            self.resource_name, 'delete')

    def test_delete_deleting_server(self):
        share_server = FakeShareServer(status=constants.STATUS_DELETING)
        self.mock_object(db_api, 'share_server_get',
                         mock.Mock(return_value=share_server))
        self.assertRaises(exc.HTTPForbidden,
                          self.controller.delete,
                          FakeRequestAdmin,
                          share_server['id'])
        policy.check_policy.assert_called_once_with(
            CONTEXT, self.resource_name, 'delete')
